local M = {}
M.type = "auxiliary"
M.relevantDevice = nil
M.defaultOrder = 1000

-- Required modules
local debugutil = require("debugutil")("auto_stop", false)

-- Constants/helpers
local supportedGearboxes = {
    cvtGearbox = true,
    automaticGearbox = true
}
local mpsToMph = 2.23694
local autoStopCoolantThreshold = 50 -- degrees Celsius
local autoStopSpeedThreshold = 0.1 -- MPH
local autoStopReenableSpeedThreshold = 3.0 -- MPH
local autoStopMinStoppedTime = 0.2 -- seconds
local engineRestartDelayTime = 0.55 -- seconds
local engineStallDetectTime = 0.5

-- Controllers and devices
local engine = nil
local gearbox = nil

-- Initial data
local initialEngineStarterTorque = 0
local initialEngineStarterMaxAV = 0
local initialEngineStarterTime = 0
local initialEngineStartCoef = 0

-- State variables
local state = "init"
local nextState = nil
local timeSinceWheelsStopped = 0
local timeSinceAutoStopped = 0
local timeSinceEngineRestart = 0
local timeSinceEngineStalled = 0
local hasMovedSinceRestart = false

-- Easy-access upvalues
local watertemp = 0
local engineRunning = false
local speedMph = 0
local gear = nil
local brake = 0

local function message(str, time)
    local _time = time or 3
    guihooks.trigger("Message", {ttl = time, msg = str, category = "fill", icon = "check_circle"})
end

local function why(str)
    dText(10, "Auto-stop disabled because " .. str)
end

local function canAutoStop()
    -- Check criteria for auto-stop

    if watertemp < autoStopCoolantThreshold then
        -- Engine must be warm enough
        why("coolant temp")
        return false
    end
    if gear ~= "D" then
        -- Must be in Drive
        why("gear")
        return false
    end
    if brake < M.brakeThreshold then
        -- Must have foot on brake
        why("brake")
        return false
    end
    if speedMph > autoStopSpeedThreshold or timeSinceWheelsStopped < autoStopMinStoppedTime then
        -- Must be stopped
        why("speed")
        return false
    end
    if timeSinceAutoStopped > M.autoStopMaxTime then
        -- Can't be auto-stopped longer than the set timeout
        why("timeout")
        return false
    end

    -- TODO: check starter battery level once implemented by BeamNG

    return true
end

local function processState(dt)
    debugutil.dText(0, "Water temp: " .. watertemp)
    debugutil.dText(1, "Gear: " .. gear)
    debugutil.dText(2, "Brake: " .. brake)
    debugutil.dText(3, "Speed: " .. speedMph)
    debugutil.dText(4, "Time auto-stopped: " .. timeSinceAutoStopped)
    debugutil.dText(5, "State: " .. state)

    if state == "init" then
        timeSinceWheelsStopped = 0
        timeSinceAutoStopped = 0
        timeSinceEngineRestart = 0
        timeSinceEngineStalled = 0
        hasMovedSinceRestart = false
        
        if engineRunning then
            nextState = "idle"
        else
            nextState = "shutdown"
        end
    elseif state == "idle" then
        -- Check if we can auto-stop

        if engineRunning then
            timeSinceEngineStalled = 0
            
            if speedMph <= autoStopSpeedThreshold then
                timeSinceWheelsStopped = math.min(timeSinceWheelsStopped + dt, 999.0)
            else
                timeSinceWheelsStopped = 0
            end

            if speedMph > autoStopReenableSpeedThreshold then
                hasMovedSinceRestart = true
            end
            
            if gear ~= "D" then
                -- Can't transition into auto-stop just by shifting to drive
                hasMovedSinceRestart = false
            end

            if canAutoStop() and hasMovedSinceRestart then
                timeSinceAutoStopped = 0
                nextState = "autostop"
            end
        else
            -- Engine stalled or was shut down by driver
            
            timeSinceEngineStalled = timeSinceEngineStalled + dt
            
            if timeSinceEngineStalled > engineStallDetectTime then nextState = "shutdown" end
        end
    elseif state == "shutdown" then
        -- See if engine is started so we can resume logic

        if engineRunning then
            nextState = "idle"
        end
    elseif state == "autostop" then
        -- Handle stopping engine and checking if we need to restart

        engine.ignitionCoef = 0 -- Kill engine
        engine.isDisabled = true -- Prevent the user from manually restarting the engine
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
        timeSinceAutoStopped = timeSinceAutoStopped + dt
        hasMovedSinceRestart = false

        if not canAutoStop() then
            nextState = "restart"
        end
    elseif state == "restart" then
        engine.isDisabled = false
        engine.starterTorque = initialEngineStarterTorque * 2
        engine.starterMaxAV = engine.idleAV * 0.5
        engine.starterThrottleKillTime = initialEngineStarterTime / 2 -- Restarting is faster
        engine.idleStartCoef = 0.85
        timeSinceEngineRestart = 0
        gearbox:setMode("neutral") -- Temporarily disengage transmission
        engine:activateStarter()

        nextState = "restarting"
    elseif state == "restarting" then
        engine.ignitionCoef = M.startingIgnitionCoef
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
        electrics.values.brake = 1.0
        
        if engineRunning and engine.starterEngagedCoef < 0.1 then
            timeSinceEngineRestart = timeSinceEngineRestart + dt
            engine.starterTorque = initialEngineStarterTorque
            engine.starterMaxAV = initialEngineStarterMaxAV
            engine.starterThrottleKillTime = initialEngineStarterTime
            
            -- Reengage transmission
            if gear == "P" then
                gearbox:setMode("park") 
            elseif gear ~= "N" then
                gearbox:setMode("drive") 
            end
            
            electrics.values.brake = brake
        end
        
        if timeSinceEngineRestart > engineRestartDelayTime then -- Restart complete
            engine.ignitionCoef = 1.0
            engine.idleStartCoef = initialEngineStartCoef
            nextState = "idle"
        end
    end
end

local function updateGFX(dt)
    watertemp = electrics.values.watertemp
    engineRunning = not (engine.isStalled or engine.ignitionCoef < 0.01)
    speedMph = (electrics.values.wheelspeed or 0) * mpsToMph
    gear = electrics.values.gear or "P"
    brake = input.brake or 0

    processState(dt)

    if nextState then
        state = nextState
        nextState = nil
    end
end

local function init(jbeamData)
    M.autoStopMaxTime = jbeamData.autoStopMaxTime or 60 -- seconds
    M.brakeThreshold = jbeamData.brakeThreshold or 0.15
    M.startingIgnitionCoef = jbeamData.startingIgnitionCoef or 0.2
end

local function initSecondStage()
    engine = powertrain.getDevice("mainEngine")
    gearbox = powertrain.getDevice("gearbox")

    if engine and gearbox and supportedGearboxes[gearbox.type or "none"] then
        initialEngineStarterTorque = engine.starterTorque
        initialEngineStarterMaxAV = engine.starterMaxAV
        initialEngineStarterTime = engine.starterThrottleKillTime
        initialEngineStartCoef = engine.idleStartCoef

        M.updateGFX = updateGFX
    end
end

local function reset()
    state = "init"
end

M.updateGFX = nop
M.init = init
M.initSecondStage = initSecondStage
M.reset = reset

return M
